-module(test).

-export([do/1]).
%%% Internal exports
-export([
	 qqq/0, qqq/2,
	 delete_all/0, delete_all/2,
	 insert_rand/0, insert_rand/2,
	 insert_3tup_rand/0, insert_3tup_rand/2,
	 insert_ref/0, insert_ref/2,
	 insert_seq/0, insert_seq/2,
	 insert_3tup_seq/0, insert_3tup_seq/2,
	 lookup_rand/0, lookup_rand/2,
	 lookup_seq/0, lookup_seq/2,
	 lookup_3tup_rand/0, lookup_3tup_rand/2,
	 lookup_3tup_seq/0, lookup_3tup_seq/2,
	 print_table/0, print_table/2,
	 rand_testdistrib/0, rand_testdistrib/2,
	 self_ps/0, self_ps/2,
	 sleep/0, sleep/2,
	 spit_rand/0, spit_rand/2,
	 update_counter_rand/0, update_counter_rand/2,
	 update_counter_seq/0, update_counter_seq/2,
	 update_forward/0, update_forward/2,
	 traverse_forward/0, traverse_forward/2,
	 traverse_backward/0, traverse_backward/2
	]).
%%% Debugging exports
-export([unatomify/1, test_defined/1]).

%%% We use single arg style for do() so that we can automate via
%%% "erl -s test2 do bag insert_seq 400" or
%%% "erl -s test2 do set insert 11000000 3 traverse_forw lookup_seq 11000000"

do(ArgList) when list(ArgList) ->
    [TabType|AList] = unatomify(ArgList),
    Tab = ets:new(t, [TabType, private]),
    Res = do2(Tab, AList),
    format_results(Res),
    Res.

do2(Tab, AList) ->
    FAndAList = parse_alist(AList),
    [{DoFunc, timer:tc(?MODULE, DoFunc, [Args, Tab])} ||
	{DoFunc, Args} <- FAndAList].

qqq() ->
    ok.
qqq([N], Tab) ->
    io:format("QQQ: value of Tab=~w at ~w = ~w\n", [Tab, N, ets:lookup(Tab, N)]).

delete_all() ->
    ok.
delete_all([], Tab) ->
    ets:safe_fixtable(Tab, true),
    V = delete_all2(ets:first(Tab), Tab),
    ets:safe_fixtable(Tab, false),
    V.
delete_all2('$end_of_table', Tab) ->
    ok;
delete_all2(Key, Tab) ->
    Next = ets:next(Tab, Key),
    ets:delete(Tab, Key),
    delete_all2(Next, Tab).

insert_ref() ->
    ok.
insert_ref([N], Tab) ->
    Fun0 = fun(X) -> true = ets:insert(Tab, {make_ref(), 1}) end,
    do_loop(N, Fun0).

print_table() ->
    ok.
print_table(_, Tab) ->
    ets:db_info(Tab, data).

rand_testdistrib() ->
    ok.
rand_testdistrib([N], Tab) ->
    rand_testdistrib(N, N div 4, make_randfun(N), Tab).
rand_testdistrib(0, Last, RandFun, Tab) ->
    io:format("	rand_testdistrib: info = ~p\n", [ets:info(Tab)]),
    ok;
rand_testdistrib(N, Last, RandFun, Tab) ->
    case ets:lookup(Tab, Last) of
	[] -> ets:insert(Tab, {Last, 1});
	_  -> ets:update_counter(Tab, Last, 1)
    end,
    New = RandFun(Last),
    rand_testdistrib(N - 1, New, RandFun, Tab).

spit_rand() ->
    ok.
spit_rand([N], _Tab) ->
    spit_rand(N, N div 4, make_randfun(N)).
spit_rand(0, Last, RandFun) ->
    ok;
spit_rand(N, Last, RandFun) ->
    New = RandFun(Last),
    io:format("prand ~w\n", [New]),
    spit_rand(N - 1, New, RandFun).

self_ps() ->
    ok.
self_ps([], _Tab) ->
    % Solaris: V = os:cmd(lists:flatten(io_lib:format("ps -ly -p ~s", [os:getpid()]))),
    V = os:cmd(lists:flatten(io_lib:format("ps u~s", [os:getpid()]))),
    io:format("~s\n", [V]),
    ok.

sleep() ->
    ok.
sleep([N], _Tab) ->
    timer:sleep(N),
    ok.

insert_rand() ->
    ok.
insert_rand([N], Tab) ->
    insert_rand([N, N], Tab);
insert_rand([Num, MaxKey], Tab) ->
    insert_rand([Num, MaxKey, 1], Tab);
insert_rand([Num, MaxKey, Val], Tab) ->
    insert_rand(Num, MaxKey div 4, make_randfun(MaxKey), Val, Tab).
insert_rand(0, Last, RandFun, Val, Tab) ->
    ok;
insert_rand(N, Last, RandFun, Val, Tab) ->
    New = RandFun(Last),
    true = ets:insert(Tab, {New, Val}),
    insert_rand(N - 1, New, RandFun, Val, Tab).

insert_3tup_rand() ->
    ok.
insert_3tup_rand([N], Tab) ->
    insert_3tup_rand([N, N], Tab);
insert_3tup_rand([Num, MaxKey], Tab) ->
    insert_3tup_rand([Num, MaxKey, 1], Tab);
insert_3tup_rand([Num, MaxKey, Val], Tab) ->
    insert_3tup_rand(Num, MaxKey div 4, make_randfun(MaxKey), Val, Tab).
insert_3tup_rand(0, Last, RandFun, Val, Tab) ->
    ok;
insert_3tup_rand(N, Last, RandFun, Val, Tab) ->
    New = RandFun(Last),
    true = ets:insert(Tab, make_3tup(New, Val)),
    insert_3tup_rand(N - 1, New, RandFun, Val, Tab).

insert_seq() ->
    ok.
insert_seq([N], Tab) ->
    insert_seq([N, 1], Tab);
insert_seq([N, Skip], Tab) ->
    insert_seq([N, Skip, 1], Tab);
insert_seq([N, Skip, Val], Tab) ->
    insert_seq([N, Skip, Val, 0], Tab);
insert_seq([N, Skip, Val, Stop], Tab) ->
    Fun0 = fun(X) -> true = ets:insert(Tab, {X, 1}) end,
    do_loop_skip(N, Skip, Stop, Fun0).

insert_3tup_seq() ->
    ok.
insert_3tup_seq([N], Tab) ->
    insert_3tup_seq([N, 1], Tab);
insert_3tup_seq([N, Skip], Tab) ->
    insert_3tup_seq([N, Skip, 1], Tab);
insert_3tup_seq([N, Skip, Val], Tab) ->
    insert_3tup_seq([N, Skip, Val, 0], Tab);
insert_3tup_seq([N, Skip, Val, Stop], Tab) ->
    Fun0 = fun(X) -> T = make_3tup(X, Val),
		     true = ets:insert(Tab, {T, 1})
	   end,
    do_loop_skip(N, Skip, Stop, Fun0).

lookup_rand() ->
    ok.
lookup_rand([N], Tab) ->
    lookup_rand([N, N], Tab);
lookup_rand([Num, MaxKey], Tab) ->
    lookup_rand([Num, MaxKey, 1], Tab);
lookup_rand([Num, MaxKey, Val], Tab) ->
    lookup_rand(Num, MaxKey div 4, make_randfun(MaxKey), Val, Tab).
lookup_rand(0, Last, RandFun, Val, Tab) ->
    ok;
lookup_rand(N, Last, RandFun, Val, Tab) ->
    New = RandFun(Last),
    ets:lookup(Tab, New),
    lookup_rand(N - 1, New, RandFun, Val, Tab).

lookup_seq() ->
    ok.
lookup_seq([N], Tab) ->
    lookup_seq([N, 1], Tab);
lookup_seq([N, Skip], Tab) ->
    lookup_seq([N, Skip, 1], Tab);
lookup_seq([N, Skip, Val], Tab) ->
    lookup_seq([N, Skip, Val, 0], Tab);
lookup_seq([N, Skip, Val, Stop], Tab) ->
    % Fun0 = fun(X) -> [{X, 1}] = ets:lookup(Tab, X) end,
    Fun0 = fun(X) -> ets:lookup(Tab, X) end,	% Don't assume we know value!
    do_loop_skip(N, Skip, Stop, Fun0).

lookup_3tup_seq() ->
    ok.
lookup_3tup_seq([N], Tab) ->
    lookup_3tup_seq([N, 1], Tab);
lookup_3tup_seq([N, Skip], Tab) ->
    lookup_3tup_seq([N, Skip, 1], Tab);
lookup_3tup_seq([N, Skip, Val], Tab) ->
    lookup_3tup_seq([N, Skip, Val, 0], Tab);
lookup_3tup_seq([N, Skip, Val, Stop], Tab) ->
    Fun0 = fun(X) -> ets:lookup(Tab, make_3tup(X, Val)) end,
    do_loop_skip(N, Skip, Stop, Fun0).

lookup_3tup_rand() ->
    ok.
lookup_3tup_rand([N], Tab) ->
    lookup_3tup_rand([N, N], Tab);
lookup_3tup_rand([Num, MaxKey], Tab) ->
    lookup_3tup_rand([Num, MaxKey, 1], Tab);
lookup_3tup_rand([Num, MaxKey, Val], Tab) ->
    lookup_3tup_rand(Num, MaxKey div 4, make_randfun(MaxKey), Val, Tab).
lookup_3tup_rand(0, Last, RandFun, Val, Tab) ->
    ok;
lookup_3tup_rand(N, Last, RandFun, Val, Tab) ->
    New = RandFun(Last),
    ets:lookup(Tab, make_3tup(New, Val)),
    lookup_3tup_rand(N - 1, New, RandFun, Val, Tab).

traverse_forward() ->
    ok.
traverse_forward([], Tab) ->
    traverse_forward2(ets:first(Tab), Tab).
traverse_forward2('$end_of_table', Tab) ->
    ok;
traverse_forward2(Key, Tab) ->
    traverse_forward2(ets:next(Tab, Key), Tab).

traverse_backward() ->
    ok.
traverse_backward([], Tab) ->
    traverse_backward2(ets:last(Tab), Tab).
traverse_backward2('$end_of_table', Tab) ->
    ok;
traverse_backward2(Key, Tab) ->
    traverse_backward2(ets:prev(Tab, Key), Tab).

update_counter_rand() ->
    ok.
update_counter_rand([N], Tab) ->
    update_counter_rand([N, N], Tab);
update_counter_rand([Num, MaxKey], Tab) ->
    update_counter_rand([Num, MaxKey, 1], Tab);
update_counter_rand([Num, MaxKey, Incr], Tab) ->
    update_counter_rand(Num, MaxKey div 4, make_randfun(MaxKey), Incr, Tab).
update_counter_rand(0, Last, RandFun, Incr, Tab) ->
    ok;
update_counter_rand(N, Last, RandFun, Incr, Tab) ->
    New = RandFun(Last),
    ets:update_counter(Tab, New, Incr),
    update_counter_rand(N - 1, New, RandFun, Incr, Tab).

update_counter_seq() ->
    ok.
update_counter_seq([N], Tab) ->
    update_counter_seq([N, 1], Tab);
update_counter_seq([N, Skip], Tab) ->
    update_counter_seq([N, Skip, 1], Tab);
update_counter_seq([N, Skip, Val], Tab) ->
    update_counter_seq([N, Skip, Val, 0], Tab);
update_counter_seq([N, Skip, Val, Stop], Tab) ->
    Fun0 = fun(X) -> ets:update_counter(Tab, N, Val) end,
    do_loop_skip(N, Skip, Stop, Fun0).

update_forward() ->
    ok.
update_forward([Incr], Tab) ->
    update_forward(Incr, ets:first(Tab), Tab).
update_forward(Incr, '$end_of_table', Tab) ->
    ok;
update_forward(Incr, Key, Tab) ->
    ets:update_counter(Tab, Key, Incr),
    update_forward(Incr, ets:next(Tab, Key), Tab).

%%%
%%% Low-level stuff below here
%%%

do_loop(N, Fun) ->
    do_loop_skip(N, 1, 0, Fun).

do_loop_skip(N, Skip, Stop, _) when N =< Stop ->
    ok;
do_loop_skip(N, Skip, Stop, Fun) ->
    Fun(N),
    do_loop_skip(N - Skip, Skip, Stop, Fun).

%%%
%%% smaller_prime()'s algorithm stolen from erl_db.c.
%%%

smaller_prime(N) when N rem 2 == 0 ->
    smaller_prime(N - 1);
smaller_prime(N) ->
    NModI = smaller_prime_for(3, N),
    if NModI * NModI > N -> N;
       true              -> smaller_prime(N - 2)
    end.

smaller_prime_for(I, N) when I * I =< N ->
    if N rem I == 0 -> I;
       true         -> smaller_prime_for(I + 2, N)
    end;
smaller_prime_for(I, N) ->
    I.

%%% unatomify() can now convert atoms such as '{foo, bar}' and '[1, foo, 3]'

unatomify([]) ->
    [];
unatomify([H|T]) when atom(H) ->
    L = atom_to_list(H),
    [LH|LT] = L,
    if
	LH >= $0, LH =< $9 ->
	    [list_to_integer(L) | unatomify(T)];
	LH == $- ->
	    [list_to_integer(L) | unatomify(T)];
	LH == ${; LH == $[ ->
	    case erl_scan:string(L) of
		{ok, Tokens, _} ->
		    case erl_parse:parse_term(Tokens ++ [{dot, 1}]) of
			{ok, List} when list(List) -> [List | unatomify(T)];
			{ok, Term} -> [Term | unatomify(T)];
			_ -> [H | unatomify(T)]
		    end;
		_ ->
		    [H | unatomify(T)]
	    end;
	true ->
	    [H | unatomify(T)]
    end;
unatomify([H|T]) ->
    [H | unatomify(T)].    

parse_alist([]) ->
    [];
parse_alist([Func|T]) ->
    case test_defined(Func) of
	true ->
	    {Args, Rest} = parse_alist2(T),
	    [{Func, Args} | parse_alist(Rest)];
	false ->
	    throw({bad_funcname, Func})
    end.

parse_alist2(L) ->
    parse_alist2(L, []).
parse_alist2([], Acc) ->
    {lists:reverse(Acc), []};
parse_alist2([H|T] = L, Acc) ->
    case test_defined(H) of
	true ->
	    {lists:reverse(Acc), L};
	false ->
	    parse_alist2(T, [H|Acc])
    end.

test_defined(F) ->
    case catch apply(?MODULE, F, []) of
	ok ->
	    true;
	_ ->
	    false
    end.

%%% This function is fairly fast and typically hits at least 1/2 of
%%% total space from 0 through N, often more.

make_randfun(N) ->
    SmallerPrime = smaller_prime(N),
    YetSmallerPrime = smaller_prime(SmallerPrime - 1),
    fun(Old) -> (Old * SmallerPrime) rem YetSmallerPrime end.

format_results([]) ->
    ok;
format_results([H|T]) ->
    case H of
	{Test, {N, ok}} when is_integer(N) ->
	    io:format("~w ok ~w\n", [Test, N]);
	{Test, {N, Huh}} when is_integer(N) ->
	    io:format("~w FAIL ~w : ~w\n", [Test, N, Huh]);
	_ ->
	    foo
    end,
    format_results(T).

make_3tup(X, Val) ->
    case X rem 5 of
	0 -> {atom_a, X, Val};
	1 -> {foo_b, X, Val};
	2 -> {blarf_c, X, Val};
	3 -> {dog_d, X, Val};
	4 -> {eagle_e, X, Val}
    end.
